/* ************************************************************************** 
 *             The Pebbles Personal Digital Assistant Project               *
 * **************************************************************************
 * This code was written as part of the Pebbles project at                  *
 * Carnegie Mellon University.  See the following for more information:     *
 *                http://www.cs.cmu.edu/~pebbles                            *
 * This library code is available for use by others.                        *
 * Please contact us for more information: pebbles@cs.cmu.edu               *  
 *                                                                          *
 *         Copyright (C) 1998-2000 - Carnegie Mellon University             *
 * **************************************************************************/

/* 
   Pebbles library for PalmPilot.

   This file defines a set of functions that implement the app-side
   of the Pebbles protocol for a PalmPilot.

*/

#include <PalmOS.h>
#include <SerialMgrOld.h>
#include <Pebbles.h>
#include <PebblesPilot.h>
#include <PebblesResource.h>

#include "NdkLib.h"
#include <sys_socket.h>

// Interval for repeating the CHANGE_PLUGIN message, in milliseconds
#define CHANGE_PLUGIN_REPEAT_INTERVAL_MSECS     1000

// if the network fails, repeat every 5 secs
#define NETWORK_CONNECT_REPEAT_INTERVAL_SECS   (5 * sysTicksPerSecond)

// sound to play when connection is made to desktop.
#define DEFAULT_CONNECTED_SOUND		sndConfirmation

#define RECEIVE_BUFFER_SIZE 20032

#define PEBBLES_APPID 		'PEBL'
#define PEBBLES_APPVERSION 	1

typedef struct {
	char connType;
	char IPAddress[16];
} PebblesPreferenceType;


Boolean OpenNetwork(PebblesPtr pbl, char *ipAddr);
Boolean OpenSerial(PebblesPtr pbl);
void CloseNetwork(PebblesPtr pbl);
void CloseSerial(PebblesPtr pbl);
void PebblesGetPreferences(PebblesPreferenceType *prefs);
Boolean ReceiveNetworkData (PebblesPtr pbl, void *data, unsigned len, int timeout);
Boolean ReceiveSerialData (PebblesPtr pbl, void *data, unsigned len, int timeout);
unsigned char NetworkReceiveHeader (PebblesPtr pbl, unsigned *pLen, int timeout);
unsigned char SerialReceiveHeader (PebblesPtr pbl, unsigned *pLen, int timeout);


Boolean OpenNetwork(PebblesPtr pbl, char *ipAddr) {

	int retVal = 0;
	struct sockaddr_in *srv, *cli;
	SocketType *sock;
	char * ip;
	
	sock = (SocketType *) MemPtrNew(sizeof(SocketType));
	srv = (struct sockaddr_in *) MemPtrNew(sizeof(struct sockaddr_in));
	cli = (struct sockaddr_in *) MemPtrNew(sizeof(struct sockaddr_in));
	ip = MemPtrNew(StrLen(ipAddr)+1);
	StrCopy(ip, ipAddr);
	
	pbl->socket = sock;
	
	//___clear our primary data structure and initialize
	MemSet ((char*)pbl->socket, sizeof(SocketType), 0);		
	pbl->socket->NetLibTimeOut = (2 * sysTicksPerSecond);  
	pbl->socket->TCPResendTimeout = 3000;	
	pbl->socket->hostIP = ip;	
	pbl->socket->port  = PEBBLES_APP_PORTNUM;
	pbl->socket->srv_addrP = srv;
	pbl->socket->cli_addrP = cli;
	
	retVal = NdkLibGetSocket(pbl->socket, 0);
	if(retVal < 0) {
		pbl->lastErr = retVal;
		return false;
	}
	
	pbl->lastErr = 0;
	pbl->socket->mySocket = retVal;
	return true;
}

Boolean OpenSerial(PebblesPtr pbl) {
	SerSettingsType settings;
	UInt32 romVersion;
  	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);


	SysLibFind("Serial Library", &pbl->ref);
	pbl->lastErr = SerOpen(pbl->ref, 0, PEBBLES_BAUDRATE);
	if (pbl->lastErr) {
		if (pbl->lastErr == serErrAlreadyOpen)
			SerClose (pbl->ref);
		return false;
	}

	if(pbl->connType == INFRARED) {
		if(romVersion >= 0x03000000)
			SerControl(pbl->ref, serCtlIrDAEnable, 0, 0);
		else 
			// can't use IR.. default to serial
			pbl->connType = SERIAL;
	}
	
	// set port settings
	pbl->lastErr = SerGetSettings (pbl->ref, &settings);
	settings.baudRate = PEBBLES_BAUDRATE;
	settings.ctsTimeout = serDefaultCTSTimeout;
	settings.flags = serDefaultSettings;
	pbl->lastErr = SerSetSettings(pbl->ref, &settings);
	
	// increase the receive buffer
  	pbl->receiveBuffer = MemPtrNew (RECEIVE_BUFFER_SIZE);
  	ErrFatalDisplayIf (pbl->receiveBuffer == NULL, "Trying to increase receive buffer: Out of dynamic memory");
  	SerSetReceiveBuffer (pbl->ref, pbl->receiveBuffer, RECEIVE_BUFFER_SIZE);
  
  	return true;
}

void PebblesGetPreferences(PebblesPreferenceType *prefs) {
	UInt16 prefSize = sizeof(PebblesPreferenceType);

	if (PrefGetAppPreferences (PEBBLES_APPID, 0, prefs, &prefSize, true) 
		== noPreferenceFound) {
		// we didn't get prefs, fill in default values
		prefs->connType = SERIAL;
		
		MemSet(prefs->IPAddress, 16, 0);
		PrefSetAppPreferences (PEBBLES_APPID, 0, PEBBLES_APPVERSION, prefs, sizeof (PebblesPreferenceType), true);
	}
}


/***********************************************************************
 *
 * FUNCTION:    PebblesOpen
 *
 * DESCRIPTION: Open a connection to the desktop.
 *
 * PARAMETERS:  pbl -- connection state (caller must allocate)
 *              plugin -- name of plugin to open on desktop.
 *                        PebblesOpen saves a pointer to this
 *						  string, so it must live as long as the
 *                        connection.  Can be NULL, in which case
 *                        PebblesOpen doesn't attempt to open a plugin.
 *
 * RETURNED:    always returns true.
 *
 *        Note that a successful return from this function means only
 * 		  that the Pilot serial port is open.  It does not mean that
 *        the Pilot is connected to a PC with a running PebblesPC
 *        process with the desired plugin.  PebblesHandshake() or
 *        PebblesReceiveHeader() must be called periodically to watch
 *        for the desktop plugin to appear on the other end of the
 *        connection.
 *
 ***********************************************************************/

Boolean PebblesOpen (PebblesPtr pbl, char *plugin)
{
	PebblesPreferenceType prefs;
	
	PebblesGetPreferences(&prefs);

	// initialize pbl
	pbl->userName[0] = '\0'; // initialize with empty name
	pbl->plugin = plugin;
	pbl->ref = 0;
	pbl->connected = false;
	pbl->pluginConnectError = false;
	pbl->lastErr = 0;
	pbl->sound_on_connected = DEFAULT_CONNECTED_SOUND;
	pbl->connectionCallback = NULL;
	PebblesGetUserName(pbl->userName); // try to get the user name
	pbl->userNameLength = StrLen(pbl->userName);
	pbl->socket = NULL;
	pbl->connType = (PebblesConnectionType) prefs.connType;
	
	// open serial port
	
	if(pbl->connType == NETWORK) {
		if(!OpenNetwork(pbl, prefs.IPAddress)) {
			pbl->repeatDeadline = TimGetTicks() + NETWORK_CONNECT_REPEAT_INTERVAL_SECS;
			return true;	
		}
	}
	else {
		if(!OpenSerial(pbl))
			return true;
	}
	
	// connect to the plugin
	if (plugin != NULL)
		PebblesChangePlugin (pbl, plugin);

	return true;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesOpenAndWait
 *
 * DESCRIPTION: Open a connection to the desktop and waits to see if it succeeds
 *
 * PARAMETERS:  pbl -- connection state (caller must allocate)
 *              plugin -- name of plugin to open on desktop.
 *                        PebblesOpen saves a pointer to this
 *						  string, so it must live as long as the
 *                        connection.  Can be NULL, in which case
 *                        PebblesOpen doesn't attempt to open a plugin.
 *
 * RETURNED:    true if connection opened succesfully;
 *				false if error (error value available in pbl->lastErr).
 *
 *        Note that a successful return from this function means that the open succeeded and
 *	   a handshake succeeded.
 *
 ***********************************************************************/
Boolean PebblesOpenAndWait (PebblesPtr pbl, char *plugin)
{
	if (PebblesOpen(pbl, plugin)) {
	
		if (PebblesHandshake(pbl, 100)) return true;
		else {
			PebblesClose(pbl);
			return false;
			}

	}
	else {
		return false;
	}
}

/***********************************************************************
 *
 * FUNCTION:    PebblesOpenAndWaitSetSound
 *
 * DESCRIPTION: Just like  PebblesOpenAndWait, but sets the sound for connection opening
 *
 * PARAMETERS:  pbl -- connection state (caller must allocate)
 *              		plugin -- name of plugin to open on desktop. (see PebblesOpen)
 *				sound -- sound to make on connection, 0 = no sound, default = sndConfirmation
 *
 * RETURNED:    true if connection opened succesfully;
 *				false if error (error value available in pbl->lastErr).
 *
 *        Note that a successful return from this function means that the open succeeded and
 *	   a handshake succeeded.
 *
 ***********************************************************************/
Boolean PebblesOpenAndWaitSetSound (PebblesPtr pbl, char *plugin, int sound)
{

	PebblesSetConnectionSound(pbl, sound);
	return PebblesOpenAndWait(pbl, plugin);
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSetConnectionSound
 *
 * DESCRIPTION: Sets the sound that plays on opening
 *
 * PARAMETERS:  pbl -- connection state 
 *				sound -- sound to make on connection, 0 = no sound, default = sndConfirmation
 *
 * RETURNED:    none
 *
 *        Note that this must be called after PebblesOpen but before PebblesReceive or PebblesHandshake
 *		See also PebblesOpenAndWaitSetSound which does this automatically
 *
 ***********************************************************************/
void PebblesSetConnectionSound(PebblesPtr pbl,  int sound) {
	if (pbl) pbl->sound_on_connected = sound;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesClose
 *
 * DESCRIPTION: Closes connection to the desktop.
 *
 * PARAMETERS:  pbl	-- connection state
 *
 * RETURNED:    nothing
 *
 ***********************************************************************/

void CloseNetwork(PebblesPtr pbl) {
	if(pbl->socket != NULL) {
		NdkLibCloseSocket (pbl->socket);
		MemPtrFree(pbl->socket->hostIP);
		MemPtrFree(pbl->socket->srv_addrP);
		MemPtrFree(pbl->socket->cli_addrP);
		MemPtrFree(pbl->socket);
	}
}

void CloseSerial(PebblesPtr pbl) {
	if(pbl->ref != NULL) {
		// wait until outgoing bytes are sent
		SerSendWait(pbl->ref, -1);
		SysTaskDelay (1);	// SerSendWait doesn't seem to wait long enough

		SerSetReceiveBuffer (pbl->ref, NULL, 0);
		if(pbl->receiveBuffer)
			MemPtrFree(pbl->receiveBuffer);

		if(pbl->connType == INFRARED)
			SerControl(pbl->ref, serCtlIrDADisable, 0, 0); 
		
		SerClose (pbl->ref);
	}
}

void PebblesClose (PebblesPtr pbl)
{
	// tell Pebbles that we're going away
	PebblesChangePlugin (pbl, NULL);

	if(pbl->connType == NETWORK)
		CloseNetwork(pbl);
	else
		CloseSerial(pbl);

	PebblesEnableAutoPowerOff();
	pbl->connected = false;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesChangePlugin
 *
 * DESCRIPTION: Open a new plugin on the desktop (closing the
 *              current plugin, if any).
 *
 * PARAMETERS:  pbl -- connection state
 *              plugin -- name of plugin to open on desktop.
 *                        PebblesOpen saves a pointer to this
 *						  string, so it must live as long as the
 *                        connection.  Can be NULL, in which case
 *                        the current plugin is just closed.
 *
 * RETURNED:    true if request sent successfully;
 *              false if error (error value available in pbl->lastErr)
 *
 ***********************************************************************/

Boolean PebblesChangePlugin (PebblesPtr pbl, char *plugin)
{
    Boolean result;
    Boolean wasConnected;
    Boolean flowControl = (pbl->connType == SERIAL);
	unsigned pluglen = (plugin != NULL) ? StrLen (plugin) : 0;
//	CharPtr sendstring;
//	unsigned len;

	// fool PebblesSend into thinking we're connected
	wasConnected = pbl->connected;
	pbl->connected = true;
//	len = pluglen + userNameLength + 1;
//	sendstring = new char[pluglen + userNameLength];
//	strncpy(sendstring,plugin,pluglen);
//	sendstring[pluglen] = '\0';

	result = PebblesSendHeader(pbl,CMD_PEBBLES_CHANGE_PLUGIN,
					pluglen+pbl->userNameLength+3);
	result = result && PebblesSendData(pbl,plugin,pluglen);
	result = result && PebblesSendData(pbl,"\0",1);
	result = result && PebblesSendData(pbl,pbl->userName,pbl->userNameLength);
	result = result && PebblesSendData(pbl,"\0",1);
	result = result && PebblesSendData(pbl, &flowControl, 1);
	
//	strncpy((CharPtr)(sendstring[pluglen+1]),username,userlen);
//	result = PebblesSend (pbl, CMD_PEBBLES_CHANGE_PLUGIN, sendstring, len);
//	result = result && PebblesSend (pbl, CMD_PEBBLES_CHANGE_USERNAME, username, userlen);
	pbl->connected = wasConnected;
	pbl->plugin = plugin;
	pbl->repeatDeadline = TimGetTicks() + CHANGE_PLUGIN_REPEAT_INTERVAL_MSECS/10;
	return result;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSetConnectionCallback
 *
 * DESCRIPTION: Set a callback function to be called every time a
 *              new connection is established with PebblesPC (after
 *              CMD_PEBBLES_ACK_CHANGE_PLUGIN is received).
 *
 * PARAMETERS:  pbl -- connection state
 *              callback -- callback function.  May be NULL if no
 *                  callback is desired (the default).
 *
 *
 ***********************************************************************/

void PebblesSetConnectionCallback (PebblesPtr pbl, void (*callback) (PebblesPtr pbl)) 
{
    pbl->connectionCallback = callback;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesHandshake
 *
 * DESCRIPTION: Checks whether the desktop plugin is connected and
 *              available.
 *
 *				Apps should periodically call either this function
 *				or PebblesReceiveHeader to watch for handshaking
 *              requests from the desktop.  Note that this function eats
 *				all incoming messages, so an app which is expecting
 *				asynchronous messages should use PebblesReceiveHeader
 *              instead.
 *
 * PARAMETERS:  pbl -- connection state
 *				timeout -- interbyte timeout, in ticks (.01 sec).
 *						   0 means poll, -1 means wait forever.
 *
 * RETURNED:    true if plugin connected;
 *              false if plugin not connected yet or error.
 *
 ***********************************************************************/


Boolean PebblesHandshake (PebblesPtr pbl, int timeout)
{
	unsigned len;
	PebblesReceiveHeader (pbl, &len, timeout);
	return pbl->connected;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSend
 *
 * DESCRIPTION: Send a message to the desktop plugin.
 *
 * PARAMETERS:  pbl -- connection state
 *              cmd -- user-defined message code (at least CMD_PEBBLES_BASE)
 *              data -- pointer to data to send; may be NULL if len=0
 *              len -- length of data block, may be 0
 *
 * RETURNED:    true if message successfully sent to plugin;
 *              false if no plugin connected or error.
 *
 ***********************************************************************/

Boolean PebblesSend (PebblesPtr pbl, unsigned char cmd, void *data, unsigned len)
{
	return PebblesSendHeader (pbl, cmd, len)
			&& PebblesSendData (pbl, data, len);
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSendHeader
 *
 * DESCRIPTION: Send a message header to the plugin.
 *              Caller is responsible for finishing the message
 *              by sending <len> bytes of data using PebblesSendData().
 *
 * PARAMETERS:  pbl -- connection state
 *              cmd -- user-defined message code (at least CMD_PEBBLES_BASE)
 *              len -- length of data
 *
 * RETURNED:    true if message header successfully sent to plugin;
 *              false if no plugin connected or error.
 *
 ***********************************************************************/

Boolean PebblesSendHeader (PebblesPtr pbl, unsigned char cmd, unsigned len)
{
	unsigned char hdr[PEBBLES_HEADER_SIZE];

	if (!pbl->connected)
		return false;		// no hope of delivering this message

	PEBBLES_SET_CMD (hdr, cmd);
	PEBBLES_SET_LEN (hdr, len);
	
	if(pbl->connType == NETWORK) {
		if(pbl->socket->mySocket == NULL) return false;
		pbl->socket->NetLibTimeOut = 5;
		pbl->socket->data_size = sizeof(hdr);
		pbl->socket->data_buffP = (char *) hdr;
		pbl->lastErr = NdkLibWriteSocket (pbl->socket);
		if(pbl->socket->errNo == netErrSocketClosedByRemote)
			pbl->connected = false;
	}
	else {
		if(pbl->ref == NULL) return false;
		SerSend (pbl->ref, hdr, sizeof(hdr), &pbl->lastErr);
	}
	return !pbl->lastErr;
}


/***********************************************************************
 *
 * FUNCTION:    PebblesSendData
 *
 * DESCRIPTION: Send message data to the plugin.  The message should
 *              have already been started by PebblesSendHeader.
 *              PebblesSendData can be called several times to send
 *              the complete data field of the message (as long as
 *              the total data sent equals the value passed to
 *              PebblesSendHeader).
 *
 * PARAMETERS:  pbl -- connection state
 *              data -- pointer to data to send
 *              len -- length of data block, may be 0
 *
 * RETURNED:    true if data successfully sent to plugin;
 *              false if no plugin connected or error.
 *
 ***********************************************************************/

Boolean PebblesSendData (PebblesPtr pbl, void *data, unsigned len)
{
	if (!pbl->connected)
		return false;	// no hope of delivering

	if(pbl->connType == NETWORK) {
		if(pbl->socket->mySocket == NULL) return false;
		pbl->socket->NetLibTimeOut = 5;
		pbl->socket->data_size = len;
		pbl->socket->data_buffP = (char *) data;
		pbl->lastErr = NdkLibWriteSocket (pbl->socket);
		if(pbl->socket->errNo == netErrSocketClosedByRemote)
			pbl->connected = false;
	}
	else {
		if(pbl->ref == NULL) return false;
		SerSend (pbl->ref, data, len, &pbl->lastErr);
	}
	return !pbl->lastErr;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesReceiveHeader
 *
 * DESCRIPTION: Receives a message header from the desktop plugin.
 *              Depending on the timeout parameter, this function acts
 *              as either a nonblocking or a blocking read.
 *              Returns only user-defined messages (not system protocol
 *              messages).
 *
 *              If a message is returned, the caller is responsible
 *              for reading <*pLen> bytes from the serial port using
 *              PebblesReceiveData ().
 *
 * PARAMETERS:  pbl -- connection state
 *              pLen -- variable to receive length of incoming message's
 *              		data block
 *				timeout -- interbyte timeout in ticks (.01 sec)
 * 				           0 means poll, -1 means wait forever
 *
 * RETURNED:    message code of received message;
 *              0 if no message is received or error occured.
 *
 ***********************************************************************/

unsigned char NetworkReceiveHeader (PebblesPtr pbl, unsigned *pLen, int timeout)
{

	unsigned char hdr[PEBBLES_HEADER_SIZE];
	unsigned char cmd;
	int status;
	
	if(pbl->socket->mySocket == NULL)
		goto fail;
	
	//check to see if any data to be received
	pbl->socket->NetLibTimeOut = timeout;
	status = NdkLibReadSelect(pbl->socket);
	if(status) {
		// nothing there
		goto fail;
	}
	
	//receive the command	
	pbl->socket->data_size = PEBBLES_HEADER_SIZE;
	pbl->socket->data_buffP = (char *) (hdr);	  	
	status = NdkLibReadSocket (pbl->socket);
	if(status) {
		// we couldn't receive a header in time
		goto fail;
	}	
		
	
	// parse command	
	switch (cmd = hdr[0]) {
		case CMD_PEBBLES_ACK_CHANGE_PLUGIN:
			// PebblesPC acknowledged our CMD_PEBBLES_CHANGE_PLUGIN,
			// so we're safely connected to a plugin
			
			pbl->connected = true;
			if (pbl->sound_on_connected)
				SndPlaySystemSound ((SndSysBeepType)pbl->sound_on_connected);
         
            if (pbl->connectionCallback)
                pbl->connectionCallback (pbl);
			break;
		case CMD_PEBBLES_NO_ACK_CHANGE_PLUGIN:
			
			if(!pbl->pluginConnectError && FrmVisible(FrmGetActiveForm())) {
				pbl->pluginConnectError = true;
				pbl->repeatDeadline = MAX_ULONG;
				FrmCustomAlert(PluginNotFoundAlert, pbl->plugin, NULL, NULL);
			}
			break;
		case CMD_PEBBLES_STARTUP:
			// PebblesPC just started up; send it the
			// CMD_PEBBLES_CHANGE_PLUGIN message it probably missed
			PebblesChangePlugin (pbl, pbl->plugin);
			break;

		default:
			if (pbl->connected && cmd >= CMD_PEBBLES_BASE && cmd <= CMD_PEBBLES_MAX) {
				// user-defined message -- return it
				*pLen = PEBBLES_LEN (hdr);
				return cmd;
			}
			break;
	}

fail:	// no message waiting

	if(pbl->socket->errNo == netErrNotOpen) {
		pbl->connected = false;
	}

    if (!pbl->connected 
        && pbl->plugin != NULL
        && TimGetTicks () >= pbl->repeatDeadline) {
        
        CloseNetwork(pbl);
        PebblesOpen(pbl, pbl->plugin);
    }
	return 0;
}

unsigned char SerialReceiveHeader (PebblesPtr pbl, unsigned *pLen, int timeout)
{
	unsigned char hdr[PEBBLES_HEADER_SIZE];
	unsigned char cmd;
	UInt32 n;

	if(pbl->ref == NULL)
		return 0;

	pbl->lastErr = SerReceiveCheck (pbl->ref, &n);
	if (pbl->lastErr == serErrLineErr) {
		SerClearErr (pbl->ref); // clear the error
		goto fail;
	}

	while (n >= PEBBLES_HEADER_SIZE || timeout > 0) {
		// wait for full header to arrive
		pbl->lastErr = SerReceiveWait (pbl->ref, sizeof(hdr), timeout);
		if (pbl->lastErr) {
			if (pbl->lastErr == serErrLineErr) {
				SerClearErr (pbl->ref); // clear the error
			}
			goto fail;
		}

		SerReceive (pbl->ref, hdr, 1, 0, &pbl->lastErr);
		n -= 1;

		switch (cmd = hdr[0]) {
		case CMD_PEBBLES_ACK_CHANGE_PLUGIN:
			SerReceive (pbl->ref, hdr+1, PEBBLES_HEADER_SIZE-1, 0, &pbl->lastErr);
			n -= PEBBLES_HEADER_SIZE-1;
			// PebblesPC acknowledged our CMD_PEBBLES_CHANGE_PLUGIN,
			// so we're safely connected to a plugin
			pbl->connected = true;
			if (pbl->sound_on_connected)
				SndPlaySystemSound ((SndSysBeepType)pbl->sound_on_connected);
         
            if (pbl->connectionCallback)
                pbl->connectionCallback (pbl);
			break;
		case CMD_PEBBLES_NO_ACK_CHANGE_PLUGIN:
			SerReceive (pbl->ref, hdr+1, PEBBLES_HEADER_SIZE-1, 0, &pbl->lastErr);
			n -= PEBBLES_HEADER_SIZE-1;
			if(!pbl->pluginConnectError && FrmVisible(FrmGetActiveForm())) {
				pbl->pluginConnectError = true;
				FrmCustomAlert(PluginNotFoundAlert, pbl->plugin, NULL, NULL);
			}
			break;
		case CMD_PEBBLES_STARTUP:
			SerReceive (pbl->ref, hdr+1, PEBBLES_HEADER_SIZE-1, 0, &pbl->lastErr);
			n -= PEBBLES_HEADER_SIZE-1;
			// PebblesPC just started up; send it the
			// CMD_PEBBLES_CHANGE_PLUGIN message it probably missed
			PebblesChangePlugin (pbl, pbl->plugin);
			break;

		default:
			if (pbl->connected && cmd >= CMD_PEBBLES_BASE && cmd <= CMD_PEBBLES_MAX) {
				SerReceive (pbl->ref, hdr+1, PEBBLES_HEADER_SIZE-1, 0, &pbl->lastErr);
				n -= PEBBLES_HEADER_SIZE-1;
				// user-defined message -- return it
				*pLen = PEBBLES_LEN (hdr);
				return cmd;
			}
			break;
		}
	}

fail:	// no message waiting

    if (!pbl->connected 
        && pbl->plugin != NULL
        && TimGetTicks () >= pbl->repeatDeadline)
        // Still not connected -- repeat the CHANGE_PLUGIN message
        PebblesChangePlugin (pbl, pbl->plugin);

	return 0;
}

unsigned char PebblesReceiveHeader (PebblesPtr pbl, unsigned *pLen, int timeout) {
	
	if(pbl->connType == NETWORK) {
		return NetworkReceiveHeader(pbl, pLen, timeout);
	}
	else {
		return SerialReceiveHeader(pbl, pLen, timeout);
	}

}

/***********************************************************************
 *
 * FUNCTION:    PebblesReceiveData
 *
 * DESCRIPTION: Receives a message's data block.  May be called
 *              several times to receive an entire message, if
 *              desired.
 *
 * PARAMETERS:  pbl -- connection state
 *              data -- pointer to buffer for received data block.
 *                      May be NULL, in which case the bytes are
 *                      read and discarded.
 *              len -- length of data block, may be 0
 *				timeout -- interbyte timeout in ticks, i.e. 1/100s of
 * 				           a second. (0 means poll, -1 means forever)
 *
 * RETURNED:    True if data received successfully; false if error.
 *              When false is returned, some of the data may have
 *              been read.
 *
 ***********************************************************************/

Boolean ReceiveNetworkData (PebblesPtr pbl, void *data, unsigned len, int /* timeout */) {

	char *junk;
	int status;

	char tmpBuf[50];
	
	if(pbl->socket->mySocket == NULL)
		return false;

	pbl->socket->NetLibTimeOut = -1; //timeout;
		
	if(data != NULL) {

		pbl->socket->data_size = len;
		pbl->socket->data_buffP = data;
	
		status = NdkLibReadSocket (pbl->socket);
		if(status) {
			StrPrintF(tmpBuf, "Error no: %d, len: %d", (int) pbl->socket->errNo, status);
			WinDrawChars(tmpBuf, StrLen(tmpBuf), 5, 40);
		}
		
		if(pbl->socket->data_size != len) {
			StrPrintF(tmpBuf, "Error no: %x! Rec %d of %d", pbl->socket->errNo, pbl->socket->data_size, len);
			WinDrawChars(tmpBuf, StrLen(tmpBuf), 5, 40);
			status = 1;
		}
		
	}
	else {
		junk = MemPtrNew(len);
		pbl->socket->data_size = len;
		pbl->socket->data_buffP = junk;
		  	
		status = NdkLibReadSocket (pbl->socket);
		
		MemPtrFree(junk);
	}
	
	return !status;
}

Boolean ReceiveSerialData (PebblesPtr pbl, void *data, unsigned len, int timeout)
{
	if(pbl->ref == NULL)
		return false;

	if (data != NULL)
		// try to read len bytes
		SerReceive (pbl->ref, data, len, timeout, &pbl->lastErr);
	else {
		// read and discard len bytes
		unsigned char junk;
		unsigned i;

		for (i=0; i<len; ++i)
			if (SerReceive (pbl->ref, &junk, 1, timeout, &pbl->lastErr) < 1)
				break;
	}

	if (pbl->lastErr == serErrLineErr)
		SerClearErr (pbl->ref); // clear the error

	return !pbl->lastErr;
}

Boolean PebblesReceiveData (PebblesPtr pbl, void *data, unsigned len, int timeout) {
	
	if(pbl->connType == NETWORK)
		return ReceiveNetworkData(pbl, data, len, timeout);
	else
		return ReceiveSerialData(pbl, data, len, timeout);

}

/***********************************************************************
 *
 * FUNCTION:    PebblesReceiveFlush
 *
 * DESCRIPTION: Throws away any characters waiting to be read.
 *              Might be used to discard a garbled message.
 *
 * PARAMETERS:  pbl -- connection state
 *
 * RETURNED:    nothing
 *
 ***********************************************************************/

void PebblesReceiveFlush (PebblesPtr pbl)
{
	if(pbl->connType != NETWORK) {
		SerReceiveFlush (pbl->ref, 1);
	}
}


/***********************************************************************
 *
 * FUNCTION:    PebblesGetUserName
 *
 * DESCRIPTION: Gets the Pilot's user name. 
 *
 * PARAMETERS:  none
 *
 * RETURNED:    CharPtr of UserName
 *
 ***********************************************************************/

// copies user name into nameptr which should be of size dlkMaxUserNameLength + 1
void PebblesGetUserName (char *nameptr)
{
    /* Try and use the user's name */
    DlkGetSyncInfo(NULL, NULL, NULL, nameptr, NULL, NULL);
}

/***********************************************************************
 *
 * FUNCTION:    PebblesDisableAutoPowerOff
 *
 * DESCRIPTION: Disables the pilot's auto power off function and
 *				stores the user's power off preference.
 *
 * PARAMETERS:  none
 *
 * RETURNED:    nothing
 *
 ***********************************************************************/


static int 		powerOffDelay 	= -1;
static Boolean	disablePowerOff = false;

void PebblesDisableAutoPowerOff() {
	powerOffDelay = SysSetAutoOffTime(0);
}

void PebblesEnableAutoPowerOff() {
	if(powerOffDelay != -1) SysSetAutoOffTime(powerOffDelay);
}

// you can pass a PebblesPtr of NULL if you don't need to reconnect after
// closing the dialog.
Boolean PebblesConnectionFormHandleEvent(PebblesPtr pbl, EventPtr event) {

  Boolean handled = false;
  FormPtr frm;
  unsigned int connCtrl;
  FieldPtr ipFld;
  char *ip = NULL;
  MemHandle h = NULL;
  PebblesPreferenceType prefs;
  char *tmpPlugin; 
  
  frm = FrmGetActiveForm ();
  ipFld = FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, ConnectionIPAddressField));
  	  
  switch (event->eType) {
    case frmOpenEvent:
      PebblesGetPreferences(&prefs);
      
      // set the ip address 
      h = MemHandleNew(16);
      ip = MemHandleLock(h);
      StrCopy(ip, prefs.IPAddress);
      MemHandleUnlock(h);
      
      FldSetTextHandle(ipFld, h);
	  
	  ipFld->attr.underlined = 1; 
      // draw the form
      FrmDrawForm(frm);
      
      // check the correct connection box
      connCtrl = (ConnectionSerialCheckbox+prefs.connType);
      FrmSetControlGroupSelection(frm,ConnectionFormGroupID,connCtrl);
    
      handled = true;
      break;
    case ctlSelectEvent:
      switch(event->data.ctlSelect.controlID) {
        case ConnectionOKButton: 
      		
      		// copy ip addr into preferences structure
      		ip = FldGetTextPtr(ipFld);
	      	if(ip == NULL) MemSet(prefs.IPAddress, 16, 0);
	      	else StrCopy(prefs.IPAddress, ip);
	      		
      		if(FrmGetControlValue(frm, FrmGetObjectIndex(frm, ConnectionSerialCheckbox))) {
	      		prefs.connType = SERIAL;
	      	}
	      	else if(FrmGetControlValue(frm, FrmGetObjectIndex(frm, ConnectionInfraredCheckbox))) {
	      		prefs.connType = INFRARED;
	      	}
      		else if(FrmGetControlValue(frm, FrmGetObjectIndex(frm, ConnectionNetworkCheckbox))) {
      			prefs.connType = NETWORK;
	      			// can't connect to the network if there isn't an address
	      			if(ip == NULL || StrLen(ip) == 0) {
	      				SndPlaySystemSound(sndError);
	      				return true;
	      			}
      		}
      		else {
      			prefs.connType = SERIAL;	
      		}	
	      	// set preferences
			PrefSetAppPreferences (PEBBLES_APPID, 0, PEBBLES_APPVERSION, &prefs, sizeof (PebblesPreferenceType), true);
			
			if(pbl != NULL) {
				tmpPlugin = pbl->plugin;
		      	// close old Pebbles structure
		      	PebblesClose(pbl);
		      	
		      	// open new connection
		      	pbl->connType = (PebblesConnectionType) prefs.connType;
				PebblesOpen(pbl, tmpPlugin);
			}
	      
        case ConnectionCancelButton:
	      	h = FldGetTextHandle(ipFld);
	      	FldSetTextHandle(ipFld, NULL);
	      	MemHandleFree(h);
	      	
	        // returns to the window that called FrmPopupForm()
	        FrmReturnToForm(0);
	        // handle the frmUpdateEvent if you need to refresh the display
	        FrmUpdateForm(FrmGetActiveFormID(), 0);
	        handled = true;
	        break;
	        
	    default:
	    	break;
      }
      break;
    default:
      break;
  }  
  return handled;
}

